/**
 * @file main.cxx
 * @brief The main entry point for the application.
 */

#include <limits>
#include <iostream>
#include <fstream>
#include <string.h>
#include <time.h>

float fmi = std::numeric_limits<float>::max();
float fma = std::numeric_limits<float>::min();

// std::
#define STD std::
// print something onto screen
#define PRINTLN(v) STD cout << v << STD endl;
// square
#define sqr(v) ((v)*(v))
// min
#define MIN(a,b) ((a) < (b)) ? (a) : (b)
// max
#define MAX(a,b) ((a) > (b)) ? (a) : (b)

#if 0
// convert a single vector float-type data into unsigned char data
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	char ec = 'X';
	unsigned char * ubuffer;
	unsigned char uc;
	char * bufferX = NULL;
	char * bufferY = NULL;
	char * bufferZ = NULL;
	char * pbufX = NULL, * pbufY = NULL, * pbufZ = NULL;
	unsigned char * ubuf = NULL;
	float fx, fy, fz, f;
	float fmin = 0;
	float fmax = (float) sqrt(sqr(0.5f) + sqr(0.5f) + sqr(0.5f));
	int x, y, z, i;
	int nr = 0;
	char filenameX[64], filenameY[64], filenameZ[64];
	char outfilename[64];
	char hdoutfilename[64];
	bool convertBE = false;
	if (argc < 5) {
		PRINTLN(argv[0] << " [options] infilex infiley infilez outfile");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -vmin=v  minimal velocity (default 0)");
		PRINTLN("   -vmax=v  maximal velocity (default 0.866)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN(" infile[xyz]: input file name  (e.g. bin/TS21z_X_R2_004500.bin)");
		PRINTLN(" outfile    : output file name (e.g. vox/TS21z_V_R2_004500.vox)");
		return 0;
	}
	x = y = z = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
			if (p[1] == 'x') X = atoi(&p[3]);
			if (p[1] == 'y') Y = atoi(&p[3]);
			if (p[1] == 'z') Z = atoi(&p[3]);
			if (p[1] == 'v') {
				if (p[2] == 'm' && p[3] == 'i' && p[4] == 'n') fmin = (float) atof(&p[6]);
				if (p[2] == 'm' && p[3] == 'a' && p[4] == 'x') fmax = (float) atof(&p[6]);
			}
			if (p[1] == 'c') convertBE = true;
		} else {
			if (x == -1) { x = 0; strcpy(filenameX, p); }
			else if (y == -1) { y = 0; strcpy(filenameY, p); }
			else if (z == -1) { z = 0; strcpy(filenameZ, p); }
			else {
				strcpy(outfilename, p);
				strcpy(hdoutfilename, p);
				int l1 = (int) strlen(hdoutfilename);
				int l2 = l1-1;
				while (l2 >= 0 && l2 >= l1-4) {
					if (hdoutfilename[l2] == '.') {
						hdoutfilename[l2+1] = 'h';
						hdoutfilename[l2+2] = 'd';
						hdoutfilename[l2+3] = '\0';
						break;
					}
					l2--;
				}
				if (l2 == -1 || l2 == l1-5) strcat(hdoutfilename, ".hd");
			}
		}
	}
	PRINTLN("converting");
	STD ifstream inX(filenameX, STD ios_base::binary);
	STD ifstream inY(filenameY, STD ios_base::binary);
	STD ifstream inZ(filenameZ, STD ios_base::binary);
	STD ofstream out(outfilename, STD ios_base::binary);
	STD ofstream hdout(hdoutfilename);
	if (inX.fail())  { PRINTLN("Could not open " << filenameX); return 0; }
	if (inY.fail())  { PRINTLN("Could not open " << filenameY); return 0; }
	if (inZ.fail())  { PRINTLN("Could not open " << filenameZ); return 0; }
	if (out.fail())  { PRINTLN("Could not open " << outfilename); return 0; }
	if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
	bufferX = new char[4*X*Y];
	bufferY = new char[4*X*Y];
	bufferZ = new char[4*X*Y];
	ubuffer = new unsigned char[X*Y];
	for (z = 0; z < Z; z++) {
		inX.read(bufferX, 4*X*Y);
		inY.read(bufferY, 4*X*Y);
		inZ.read(bufferZ, 4*X*Y);
		pbufX = bufferX;
		pbufY = bufferY;
		pbufZ = bufferZ;
		ubuf = ubuffer;
		for (y = 0; y < Y; y++) {
			for (x = 0; x < X; x++) {
				if (convertBE) {
					STD swap<char>(pbufX[0], pbufX[3]); STD swap<char>(pbufX[1], pbufX[2]);
					STD swap<char>(pbufY[0], pbufY[3]); STD swap<char>(pbufY[1], pbufY[2]);
					STD swap<char>(pbufZ[0], pbufZ[3]); STD swap<char>(pbufZ[1], pbufZ[2]);
				}
				// x
				fx = *((float*) pbufX);
				pbufX += 4;
				// y
				fy = *((float*) pbufY);
				pbufY += 4;
				// z
				fz = *((float*) pbufZ);
				pbufZ += 4;
				// vector
				f = (float) sqrt(sqr(fx) + sqr(fy) + sqr(fz));
				f = MIN(fmax, MAX(fmin, f));
				uc = (unsigned char)(((f - fmin) / (fmax - fmin))*255.f);
				*ubuf = uc;
				ubuf++;
			}
		}
		out.write((const char *) ubuffer, X*Y);
	}
	hdout << "Size " << X << " " << Y << " " << Z << STD endl;
	inX.close();
	inY.close();
	inZ.close();
	out.close();
	hdout.close();
	delete [] bufferX;
	delete [] bufferY;
	delete [] bufferZ;
	delete [] ubuffer;
	return 0;
}
#endif
#if 0
// convert multiple vector float-type data into unsigned char data
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	char ec = 'X';
	unsigned char * ubuffer;
	unsigned char uc;
	char * bufferX = NULL;
	char * bufferY = NULL;
	char * bufferZ = NULL;
	char * pbufX = NULL, * pbufY = NULL, * pbufZ = NULL;
	unsigned char * ubuf = NULL;
	float fx, fy, fz, f;
	float fmin = 0;
	float fmax = (float) sqrt(sqr(0.5f) + sqr(0.5f) + sqr(0.5f));
	int x, y, z, i;
	int nr = 0;
	int startnr = 100;
	int endnr = 22700;
	int incnr = 100;
	char filenameX[64], filenameY[64], filenameZ[64];
	char formatX[64], formatY[64], formatZ[64];
	char outfilename[64];
	char outformat[64];
	char hdoutfilename[64];
	char hdoutformat[64];
	bool convertBE = false;
	if (argc < 5) {
		PRINTLN(argv[0] << " [options] informatx informaty informatz outformat");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -vmin=v  minimal velocity (default 0)");
		PRINTLN("   -vmax=v  maximal velocity (default 0.866)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN("   -start=s first number (default 100)");
		PRINTLN("   -end=s   final number (default 22700)");
		PRINTLN("   -inc=s   increment between numbers (default 100)");
		PRINTLN(" informat  : format of input file names  (e.g. bin/TS21z_X_R2_%06d.bin)");
		PRINTLN(" outformat : format of output file names (e.g. vox/TS21z_V_R2_%06d.vox)");
		return 0;
	}
	x = y = z = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
			if (p[1] == 'x') X = atoi(&p[3]);
			if (p[1] == 'y') Y = atoi(&p[3]);
			if (p[1] == 'z') Z = atoi(&p[3]);
			if (p[1] == 'v') {
				if (p[2] == 'm' && p[3] == 'i' && p[4] == 'n') fmin = (float) atof(&p[6]);
				if (p[2] == 'm' && p[3] == 'a' && p[4] == 'x') fmax = (float) atof(&p[6]);
			}
			if (p[1] == 'c') convertBE = true;
			if (p[1] == 's') startnr = atoi(&p[7]);
			if (p[1] == 'e') endnr = atoi(&p[5]);
			if (p[1] == 'i') incnr = atoi(&p[5]);
		} else {
			if (x == -1) { x = 0; strcpy(formatX, p); }
			else if (y == -1) { y = 0; strcpy(formatY, p); }
			else if (z == -1) { z = 0; strcpy(formatZ, p); }
			else {
				strcpy(outformat, p);
				strcpy(hdoutformat, p);
				int l1 = (int) strlen(hdoutformat);
				int l2 = l1-1;
				while (l2 >= 0 && l2 >= l1-4) {
					if (hdoutformat[l2] == '.') {
						hdoutformat[l2+1] = 'h';
						hdoutformat[l2+2] = 'd';
						hdoutformat[l2+3] = '\0';
						break;
					}
					l2--;
				}
				if (l2 == -1 || l2 == l1-5) strcat(hdoutformat, ".hd");
			}
		}
	}
	PRINTLN("converting");
	bufferX = new char[4*X*Y];
	bufferY = new char[4*X*Y];
	bufferZ = new char[4*X*Y];
	ubuffer = new unsigned char[X*Y];
	for (nr = startnr; nr <= endnr; nr += incnr) {
		PRINTLN(nr);
		sprintf(filenameX, formatX, nr);
		sprintf(filenameY, formatY, nr);
		sprintf(filenameZ, formatZ, nr);
		sprintf(outfilename, outformat, nr);
		sprintf(hdoutfilename, hdoutformat, nr);
		STD ifstream inX(filenameX, STD ios_base::binary);
		STD ifstream inY(filenameY, STD ios_base::binary);
		STD ifstream inZ(filenameZ, STD ios_base::binary);
		STD ofstream out(outfilename, STD ios_base::binary);
		STD ofstream hdout(hdoutfilename);
		if (inX.fail())  { PRINTLN("Could not open " << filenameX); return 0; }
		if (inY.fail())  { PRINTLN("Could not open " << filenameY); return 0; }
		if (inZ.fail())  { PRINTLN("Could not open " << filenameZ); return 0; }
		if (out.fail())  { PRINTLN("Could not open " << outfilename); return 0; }
		if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
		for (z = 0; z < Z; z++) {
			inX.read(bufferX, 4*X*Y);
			inY.read(bufferY, 4*X*Y);
			inZ.read(bufferZ, 4*X*Y);
			pbufX = bufferX;
			pbufY = bufferY;
			pbufZ = bufferZ;
			ubuf = ubuffer;
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					if (convertBE) {
						STD swap<char>(pbufX[0], pbufX[3]); STD swap<char>(pbufX[1], pbufX[2]);
						STD swap<char>(pbufY[0], pbufY[3]); STD swap<char>(pbufY[1], pbufY[2]);
						STD swap<char>(pbufZ[0], pbufZ[3]); STD swap<char>(pbufZ[1], pbufZ[2]);
					}
					// x
					fx = *((float*) pbufX);
					pbufX += 4;
					// y
					fy = *((float*) pbufY);
					pbufY += 4;
					// z
					fz = *((float*) pbufZ);
					pbufZ += 4;
					// vector
					f = (float) sqrt(sqr(fx) + sqr(fy) + sqr(fz));
					f = MIN(fmax, MAX(fmin, f));
					uc = (unsigned char)(((f - fmin) / (fmax - fmin))*255.f);
					*ubuf = uc;
					ubuf++;
				}
			}
			out.write((const char *) ubuffer, X*Y);
		}
		hdout << "Size " << X << " " << Y << " " << Z << STD endl;
		inX.close();
		inY.close();
		inZ.close();
		out.close();
		hdout.close();
	}
	delete [] bufferX;
	delete [] bufferY;
	delete [] bufferZ;
	delete [] ubuffer;
	return 0;
}
#endif
#if 0
// convert float-type data into unsigned char data
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	char ec = 'X';
	unsigned char * ubuffer = NULL;
	unsigned char uc;
	char * buffer = NULL;
	char * pbuf = NULL;
	unsigned char * ubuf = NULL;
	float f;
	float fmin = -0.5f;
	float fmax =  0.5f;
	int x, y, z, i;
	char filename[64];
	char outfilename[64];
	char hdoutfilename[64];
	bool convertBE = false;
	if (argc < 5) {
		PRINTLN(argv[0] << " [options] infile outfile");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -vmin=v  minimal value (default -0.5)");
		PRINTLN("   -vmax=v  maximal value (default  0.5)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN(" infile  : input file name  (e.g. bin/TS21z_X_R2_00100.bin)");
		PRINTLN(" outfile : output file name (e.g. vox/TS21z_V_R2_00100.vox)");
		return 0;
	}
	x = y = z = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
			if (p[1] == 'x') X = atoi(&p[3]);
			if (p[1] == 'y') Y = atoi(&p[3]);
			if (p[1] == 'z') Z = atoi(&p[3]);
			if (p[1] == 'v') {
				if (p[2] == 'm' && p[3] == 'i' && p[4] == 'n') fmin = (float) atof(&p[6]);
				if (p[2] == 'm' && p[3] == 'a' && p[4] == 'x') fmax = (float) atof(&p[6]);
			}
			if (p[1] == 'c') convertBE = true;
		} else {
			if (x == -1) { 
				x = 0; 
				strcpy(filename, p); 
			} else {
				strcpy(outfilename, p);
				strcpy(hdoutfilename, p);
				int l1 = (int) strlen(hdoutfilename);
				int l2 = l1-1;
				while (l2 >= 0 && l2 >= l1-4) {
					if (hdoutfilename[l2] == '.') {
						hdoutfilename[l2+1] = 'h';
						hdoutfilename[l2+2] = 'd';
						hdoutfilename[l2+3] = '\0';
						break;
					}
					l2--;
				}
				if (l2 == -1 || l2 == l1-5) strcat(hdoutfilename, ".hd");
			}
		}
	}
	//PRINTLN("determining min and max");
	//for (ec = 'X'; ec <= 'Z'; ec++) {
	//	nr = 0;
	//	for (i = 0; i < 227; i++) {
	//		nr += 100;
	//		PRINT(nr << ": ");
	//		sprintf(filename, "i:\\data\\terrashake\\bin\\TS21z_%.1s_R2_%06d", &ec, nr);
	//		STD ifstream in(filename, STD ios_base::binary);
	//		if (in.fail())  { PRINTLN("Could not open " << filename); return 0; }
	//		in.read(buffer, 4*X*Y*Z);
	//		pbuf = buffer;
	//		if (i == 0 && ec == 'X') {
	//			STD swap<char>(pbuf[0], pbuf[3]);
	//			STD swap<char>(pbuf[1], pbuf[2]);
	//			f = *((float*) pbuf);
	//			fmin = fmax = f;
	//		}
	//		for (z = 0; z < Z; z++) {
	//			for (y = 0; y < Y; y++) {
	//				for (x = 0; x < X; x++) {
	//					STD swap<char>(pbuf[0], pbuf[3]);
	//					STD swap<char>(pbuf[1], pbuf[2]);
	//					f = *((float*) pbuf);
	//					fmax = max(fmax, f);
	//					fmin = min(fmin, f);
	//					pbuf += 4;
	//				}
	//			}
	//		}
	//		PRINTLN("fmin=" << fmin << " fmax=" << fmax);
	//		in.close();
	//	}
	//}
	//f = max(fabs(fmin), fabs(fmax));
	//fmin = -f;
	//fmax =  f;
	//PRINTLN("Choose minimal and maximal values");
	//PRINTLN("fmin=" << fmin << " fmax=" << fmax);
	//fmin = -0.5f;
	//fmax =  0.5f;
	PRINTLN("converting");
	STD ifstream in(filename, STD ios_base::binary);
	STD ofstream out(outfilename, STD ios_base::binary);
	STD ofstream hdout(hdoutfilename);
	if (in.fail())  { PRINTLN("Could not open " << filename); return 0; }
	if (out.fail()) { PRINTLN("Could not open " << outfilename); return 0; }
	if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
	ubuffer = new unsigned char[X*Y];
	buffer = new char[4*X*Y];
	for (z = 0; z < Z; z++) {
		in.read(buffer, 4*X*Y);
		pbuf = buffer;
		ubuf = ubuffer;
		for (y = 0; y < Y; y++) {
			for (x = 0; x < X; x++) {
				if (convertBE) {
					STD swap<char>(pbuf[0], pbuf[3]); 
					STD swap<char>(pbuf[1], pbuf[2]);
				}
				f = *((float*) pbuf);
				pbuf += 4;
				f = MIN(fmax, MAX(fmin, f));
				uc = (unsigned char)(((f - fmin) / (fmax - fmin))*255.f);
				*ubuf = uc;
				ubuf++;
			}
		}
		out.write((const char *) ubuffer, X*Y);
	}
	hdout << "Size " << X << " " << Y << " " << Z << STD endl;
	in.close();
	out.close();
	hdout.close();
	delete [] buffer;
	delete [] ubuffer;
	return 0;
}
#endif
#if 0
// convert float-type data into unsigned char data
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	char ec = 'X';
	unsigned char * ubuffer = NULL;
	unsigned char uc;
	char * buffer = NULL;
	char * pbuf = NULL;
	unsigned char * ubuf = NULL;
	float f;
	float fmin = -0.5f;
	float fmax =  0.5f;
	int startnr = 100;
	int endnr = 22700;
	int incnr = 100;
	int x, y, z, i, nr;
	char filename[64];
	char format[64];
	char outfilename[64];
	char outformat[64];
	char hdoutfilename[64];
	char hdoutformat[64];
	bool convertBE = false;
	if (argc < 5) {
		PRINTLN(argv[0] << " [options] informatx informaty informatz outformat");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -vmin=v  minimal value (default -0.5)");
		PRINTLN("   -vmax=v  maximal value (default  0.5)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN("   -start=s first number (default 100)");
		PRINTLN("   -end=s   final number (default 22700)");
		PRINTLN("   -inc=s   increment between numbers (default 100)");
		PRINTLN(" informat  : format of input file names  (e.g. bin/TS21z_X_R2_%06d.bin)");
		PRINTLN(" outformat : format of output file names (e.g. vox/TS21z_X_R2_%06d.vox)");
		return 0;
	}
	x = y = z = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
			if (p[1] == 'x') X = atoi(&p[3]);
			if (p[1] == 'y') Y = atoi(&p[3]);
			if (p[1] == 'z') Z = atoi(&p[3]);
			if (p[1] == 'v') {
				if (p[2] == 'm' && p[3] == 'i' && p[4] == 'n') fmin = (float) atof(&p[6]);
				if (p[2] == 'm' && p[3] == 'a' && p[4] == 'x') fmax = (float) atof(&p[6]);
			}
			if (p[1] == 'c') convertBE = true;
			if (p[1] == 's') startnr = atoi(&p[7]);
			if (p[1] == 'e') endnr = atoi(&p[5]);
			if (p[1] == 'i') incnr = atoi(&p[5]);
		} else {
			if (x == -1) { x = 0; strcpy(format, p); }
			else {
				strcpy(outformat, p);
				strcpy(hdoutformat, p);
				int l1 = (int) strlen(hdoutformat);
				int l2 = l1-1;
				while (l2 >= 0 && l2 >= l1-4) {
					if (hdoutformat[l2] == '.') {
						hdoutformat[l2+1] = 'h';
						hdoutformat[l2+2] = 'd';
						hdoutformat[l2+3] = '\0';
						break;
					}
					l2--;
				}
				if (l2 == -1 || l2 == l1-5) strcat(hdoutformat, ".hd");
			}
		}
	}
	PRINTLN("converting");
	ubuffer = new unsigned char[X*Y];
	buffer = new char[4*X*Y];
	for (nr = startnr; nr <= endnr; nr += incnr) {
		PRINTLN(nr);
		sprintf(filename, format, nr);
		sprintf(outfilename, outformat, nr);
		sprintf(hdoutfilename, hdoutformat, nr);
		STD ifstream in(filename, STD ios_base::binary);
		STD ofstream out(outfilename, STD ios_base::binary);
		STD ofstream hdout(hdoutfilename);
		if (in.fail())  { PRINTLN("Could not open " << filename); return 0; }
		if (out.fail()) { PRINTLN("Could not open " << outfilename); return 0; }
		if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
		for (z = 0; z < Z; z++) {
			in.read(buffer, 4*X*Y);
			pbuf = buffer;
			ubuf = ubuffer;
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					if (convertBE) {
						STD swap<char>(pbuf[0], pbuf[3]); 
						STD swap<char>(pbuf[1], pbuf[2]);
					}
					f = *((float*) pbuf);
					pbuf += 4;
					f = MIN(fmax, MAX(fmin, f));
					uc = (unsigned char)(((f - fmin) / (fmax - fmin))*255.f);
					*ubuf = uc;
					ubuf++;
				}
			}
			out.write((const char *) ubuffer, X*Y);
		}
		hdout << "Size " << X << " " << Y << " " << Z << STD endl;
		in.close();
		out.close();
		hdout.close();
	}
	delete [] buffer;
	delete [] ubuffer;
	return 0;
}
#endif
#if 0
// compute divergence
int main(int argc, char **argv)
{
	//const char * filename    = "i:\\data\\terrashake\\bin\\TS21VelocityMesh_VS_R2.bin";
	//const char * outfilename = "i:\\data\\terrashake\\bin\\TS21VelocityMesh_VS_R2.nrm";
	//const char * filename    = "i:\\data\\terrashake\\bin\\TS21z_X_R2_000100";
	//const char * outfilename = "i:\\data\\terrashake\\vox\\TS21z_X_R2_000100.vox";
	const int X = 750;
	const int Y = 375;
	const int Z = 100;
	char ec = 'X';
	unsigned char uc;
	float * bufferX = new float[X*Y*Z];
	float * bufferY = new float[X*Y*Z];
	float * bufferZ = new float[X*Y*Z];
	float * buffer  = new float[X*Y*Z];
	float * pbufX, * pbufY, * pbufZ;
	float fx, fy, fz, f;
	int x, y, z, i;
	int nr = 0;
	char filenameX[64];
	char filenameY[64];
	char filenameZ[64];
	char outfilename[64];
	char hdoutfilename[64];
	int histo[256];
	nr = 0;
	float fmin = -0.0005f;//-3.0249f;
	float fmax =  0.0005f;// 3.0249f;
	const int N = 228;
	for (i = 0; i < 256; i++) histo[i] = 0;
	for (i = 0; i < N; i++) {
		nr += 100;
		PRINTLN(nr);
		sprintf(filenameX,      "i:\\data\\terrashake\\bin\\TS21z_X_R2_%06d", nr);
		sprintf(filenameY,      "i:\\data\\terrashake\\bin\\TS21z_Y_R2_%06d", nr);
		sprintf(filenameZ,      "i:\\data\\terrashake\\bin\\TS21z_Z_R2_%06d", nr);
		sprintf(outfilename,   "i:\\data\\terrashake\\div\\TS21z_R2_%06d.vox", nr);
		sprintf(hdoutfilename, "i:\\data\\terrashake\\div\\TS21z_R2_%06d.hd", nr);
		STD ifstream inX(filenameX, STD ios_base::binary);
		STD ifstream inY(filenameY, STD ios_base::binary);
		STD ifstream inZ(filenameZ, STD ios_base::binary);
		STD ofstream out(outfilename, STD ios_base::binary);
		STD ofstream hdout(hdoutfilename);
		if (inX.fail())  { PRINTLN("Could not open " << filenameX); return 0; }
		if (inY.fail())  { PRINTLN("Could not open " << filenameY); return 0; }
		if (inZ.fail())  { PRINTLN("Could not open " << filenameZ); return 0; }
		if (out.fail()) { PRINTLN("Could not open " << outfilename); return 0; }
		if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
		inX.read((char*) bufferX, sizeof(float)*X*Y*Z);
		inY.read((char*) bufferY, sizeof(float)*X*Y*Z);
		inZ.read((char*) bufferZ, sizeof(float)*X*Y*Z);
		pbufX = bufferX;
		pbufY = bufferY;
		pbufZ = bufferZ;
		for (z = 0; z < Z; z++) {
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					char * p;
					p = (char*) pbufX; STD swap<char>(p[0], p[3]); STD swap<char>(p[1], p[2]);
					p = (char*) pbufY; STD swap<char>(p[0], p[3]); STD swap<char>(p[1], p[2]);
					p = (char*) pbufZ; STD swap<char>(p[0], p[3]); STD swap<char>(p[1], p[2]);
					pbufX++;
					pbufY++;
					pbufZ++;
				}
			}
		}
		for (z = 0; z < Z; z++) {
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					if (x > 0 && x < X-1 && y > 0 && y < Y-1 && z > 0 && z < Z-1) {
						fx = bufferX[x-1+X*(y  + z   *Y)] - bufferX[x+1+X*(y+   z   *Y)];
						fy = bufferY[x  +X*(y-1+ z   *Y)] - bufferY[x  +X*(y+1+ z   *Y)];
						fz = bufferZ[x  +X*(y  +(z-1)*Y)] - bufferY[x  +X*(y  +(z+1)*Y)];
						fx /= 2;
						fy /= 2;
						fz /= 2;
						f = fx + fy + fz;
						fmi = min(f, fmi);
						fma = max(f, fma);
						buffer[x+X*(y+z*Y)] = f;
					} else
						buffer[x+X*(y+z*Y)] = 0;
				}
			}
		}
		PRINTLN(fmi << " -> " << fma);
		f = max(fabs(fmi), fabs(fma));
		fmi = -f;
		fma =  f;
		PRINTLN(fmi << " -> " << fma);
		for (z = 0; z < Z; z++) {
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					f = buffer[x+X*(y+z*Y)];
					uc = (unsigned char)(((f - fmi) / (fma - fmi))*255.f);
					histo[uc]++;
					out.write((const char *)&uc, 1);
				}
			}
		}
		for (x = 0; x < 256; x++) { PRINTLN(x << ":" << histo[x]); }
		hdout << "Size " << X << " " << Y << " " << Z << STD endl;
		inX.close();
		inY.close();
		inZ.close();
		out.close();
		hdout.close();
	}
	delete [] bufferX;
	delete [] bufferY;
	delete [] bufferZ;
	delete [] buffer;
	return 0;
}
#endif
#if 0
// compute particle movement
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	float f;
	int i, nr, N, x;
	char filename[64];
	char format[64];
	char outfilename[64];
	char outformat[64];
	char * p = NULL;
	//float fmin = -3.0249f;
	//float fmax =  3.0249f;
	float fscale = 1;
	float * fmove;
	int posx = 250;
	int posy = 250;
	int posz = 0;
	int startnr = 100;
	int endnr = 22700;
	int incnr = 100;
	bool convertBE = false;
	bool doAccu = false;
	bool doComma = true;
	bool doCr = false;
	if (argc < 3) {
		PRINTLN(argv[0] << " [options] informat outformat");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN("   -start=s first number (default 100)");
		PRINTLN("   -end=s   final number (default 22700)");
		PRINTLN("   -inc=s   increment between numbers (default 100)");
		PRINTLN("   -posx=v  x-position in the textue (default 250)");
		PRINTLN("   -posy=v  y-position in the textue (default 250)");
		PRINTLN("   -posz=v  z-position in the textue (default   0)");
		PRINTLN("   -accu    accumulate movements (default false)");
		PRINTLN("   -output={comma,cr}");
		PRINTLN("           comma: produce a comma-separated list (default)");
		PRINTLN("           cr   : produce a cariage-return-separated list (gnuplot)");
		PRINTLN(" informat  : format of input file names  (e.g. bin/TS21z_X_R2_%06d.bin)");
		PRINTLN(" outformat : format of output file names (e.g. div/TS21z_X_R2_%03d_%03d_%03d.dat)");
		return 0;
	}
	x = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
				 if (p[1] == 'x') X = atoi(&p[3]);
			else if (p[1] == 'y') Y = atoi(&p[3]);
			else if (p[1] == 'z') Z = atoi(&p[3]);
			else if (p[1] == 'c') convertBE = true;
			else if (p[1] == 's') startnr = atoi(&p[7]);
			else if (p[1] == 'e') endnr = atoi(&p[5]);
			else if (p[1] == 'i') incnr = atoi(&p[5]);
			else if (p[1] == 'p') {
				     if (p[4] == 'x') posx = atoi(&p[6]);
				else if (p[4] == 'y') posy = atoi(&p[6]);
				else if (p[4] == 'z') posz = atoi(&p[6]);
			}
			else if (p[1] == 'a') doAccu = true;
			else if (p[1] == 'o') {
				if (p[9] == 'o') { doComma = true; doCr = false; }
				else if (p[9] == 'r') { doCr = true; doComma = false; }
			}
		} else {
			if (x == -1) { x = 0; strcpy(format, p); }
			else {
				strcpy(outformat, p);
			}
		}
	}
	N = (endnr - startnr) / incnr + 1;
	fmove = new float[N];
	for (i = 0; i < N; i++) fmove[i] = 0;
	nr = 0;
	sprintf(outfilename, outformat, posx, posy, posz);
	STD ofstream out(outfilename);
	if (out.fail()) { PRINTLN("Could not open " << outfilename); return 0; }
	for (nr = startnr, i = 0; nr <= endnr; nr += incnr, i++) {
		PRINTLN(nr);
		sprintf(filename, format, nr);
		STD ifstream in(filename, STD ios_base::binary);
		if (in.fail())  { PRINTLN("Could not open " << filename); return 0; }
		in.seekg((posx+X*(posy + Y*posz))*sizeof(float));
		in.read((char *)&f, sizeof(float));
		in.close();
		p = (char*) &f; 
		if (convertBE) {
			STD swap<char>(p[0], p[3]); 
			STD swap<char>(p[1], p[2]);
		}
		fmove[i] = fscale*f;
		if (doAccu) if (i > 0) fmove[i] += fmove[i-1];
		//out << (fmove[i] - ((i>0) ? fmove[i-1] : 0));
		out << fmove[i];
		if (doComma) {
			if (nr != endnr) out << ',';
		} else out << STD endl;
	}
	out.close();
	delete [] fmove;
	return 0;
}
#endif
#if 1
// compute differences between consecutive datasets
int main(int argc, char **argv)
{
	int X = 750;
	int Y = 375;
	int Z = 100;
	float f;
	int i, j, nr, x, y, z;
	char filename[64];
	char format[64];
	char outfilename[64];
	char hdoutfilename[64];
	char outformat[64];
	char hdoutformat[64];
	char * p = NULL;
	float fmin = -0.5f;
	float fmax =  0.5f;
	float fscale = 1;
	unsigned char * ubuffer = NULL;
	unsigned char * ubuf = NULL;
	float * buffer[2];
	float * fp[2];
	int startnr = 100;
	int endnr = 22700;
	int incnr = 100;
	bool convertBE = false;
	bool doComma = true;
	bool doCr = false;
	if (argc < 3) {
		PRINTLN(argv[0] << " [options] informat outformat");
		PRINTLN(" options:");
		PRINTLN("   -x=res   x resolution (default 750)");
		PRINTLN("   -y=res   y resolution (default 375)");
		PRINTLN("   -z=res   z resolution (default 100)");
		PRINTLN("   -c       convert to big endian");
		PRINTLN("   -vmin=v  minimal value (default -0.5)");
		PRINTLN("   -vmax=v  maximal value (default  0.5)");
		PRINTLN("   -start=s first number (default 100)");
		PRINTLN("   -end=s   final number (default 22700)");
		PRINTLN("   -inc=s   increment between numbers (default 100)");
		PRINTLN(" informat  : format of input file names  (e.g. bin/TS21z_X_R2_%06d.bin)");
		PRINTLN(" outformat : format of output file names (e.g. vox/TS21z_X_R2_%06d.vox)");
		return 0;
	}
	x = -1;
	for (i = 1; i < argc; i++) {
		char * p = argv[i];
		if (p[0] == '-') {
				 if (p[1] == 'x') X = atoi(&p[3]);
			else if (p[1] == 'y') Y = atoi(&p[3]);
			else if (p[1] == 'z') Z = atoi(&p[3]);
			else if (p[1] == 'c') convertBE = true;
			else if (p[1] == 's') startnr = atoi(&p[7]);
			else if (p[1] == 'e') endnr = atoi(&p[5]);
			else if (p[1] == 'i') incnr = atoi(&p[5]);
			else if (p[1] == 'v') {
				if (p[2] == 'm' && p[3] == 'i' && p[4] == 'n') fmin = (float) atof(&p[6]);
				if (p[2] == 'm' && p[3] == 'a' && p[4] == 'x') fmax = (float) atof(&p[6]);
			}
		} else {
			if (x == -1) { x = 0; strcpy(format, p); }
			else {
				strcpy(outformat, p);
				strcpy(hdoutformat, p);
				int l1 = (int) strlen(hdoutformat);
				int l2 = l1-1;
				while (l2 >= 0 && l2 >= l1-4) {
					if (hdoutformat[l2] == '.') {
						hdoutformat[l2+1] = 'h';
						hdoutformat[l2+2] = 'd';
						hdoutformat[l2+3] = '\0';
						break;
					}
					l2--;
				}
				if (l2 == -1 || l2 == l1-5) strcat(hdoutformat, ".hd");
			}
		}
	}
	nr = 0;
	buffer[0] = new float[X*Y*Z];
	buffer[1] = new float[X*Y*Z];
	ubuffer = new unsigned char[X*Y];
	i = 0;
	fp[0] = fp[1] = NULL;
	// i is current, j is previous
	for (nr = startnr; nr <= endnr; nr += incnr) {
		PRINTLN(nr);
		sprintf(filename, format, nr);
		sprintf(outfilename, outformat, nr);
		sprintf(hdoutfilename, hdoutformat, nr);
		STD ifstream in(filename, STD ios_base::binary);
		STD ofstream out(outfilename, STD ios_base::binary);
		STD ofstream hdout(hdoutfilename);
		if (in.fail())  { PRINTLN("Could not open " << filename); return 0; }
		if (out.fail()) { PRINTLN("Could not open " << outfilename); return 0; }
		if (hdout.fail()) { PRINTLN("Could not open " << hdoutfilename); return 0; }
		j = (i+1)%2;
		in.read((char *) buffer[i], X*Y*Z*sizeof(float));
		fp[i] = buffer[i];
		fp[j] = ((nr == startnr) ? NULL : buffer[j]);
		for (z = 0; z < Z; z++) {
			ubuf = ubuffer;
			for (y = 0; y < Y; y++) {
				for (x = 0; x < X; x++) {
					if (convertBE) {
						p = (char *) fp[i];
						STD swap<char>(p[0], p[3]); 
						STD swap<char>(p[1], p[2]);
					}
					f = fscale*(*(fp[i]));
					if (fp[j]) {
						f -= *(fp[j]);
						fp[j]++;
					}
					fp[i]++;
					f = MIN(fmax, MAX(fmin, f));
					*ubuf = (unsigned char)(((f - fmin) / (fmax - fmin))*255.f);
					ubuf++;
				}
			}
			out.write((const char *) ubuffer, X*Y);
		}
		i = j;
		hdout << "Size " << X << " " << Y << " " << Z << STD endl;
		in.close();
		out.close();
		hdout.close();
	}
	delete [] buffer[0];
	delete [] buffer[1];
	delete [] ubuffer;
	return 0;
}
#endif
